#version 430

#define PI 3.141592653589

// outputs from vertex shader
in vec3 vert_world_position;
in vec3 vert_normal;
in vec2 vert_uv;
in vec4 vert_light_pos;
in vec3 light_direction;
in mat3 TBN;

layout(location = 31) uniform sampler2D texture_color;
layout(location = 32) uniform sampler2D texture_normal;
layout(location = 33) uniform sampler2D texture_metallic;
layout(location = 34) uniform sampler2D texture_roughness;
layout(location = 35) uniform sampler2D texture_ao;

layout(location = 36) uniform sampler2D depth_map;
layout(location = 37) uniform samplerCube skybox;

layout(location = 20) uniform vec4 base_color;
layout(location = 21) uniform float glow;

layout(location = 26) uniform vec3 light_dir;
layout(location = 27) uniform vec3 light_color;
layout(location = 28) uniform vec3 camera_position;

layout(location = 40) uniform float model_type;
layout(location = 42) uniform vec4 fog;

// TODO: change all these to uniform!
const vec3 ambient = vec3(0.0);
const float brightness_threshold = 1.3;
const float bloom_damping = 0.25;
const float shadow_damping = 0.6;
const float max_metallic = 1.0;

layout(location = 0) out vec4 frag_color;
layout(location = 1) out vec4 glow_color;

float shadow_calculation(vec4 pos_light_space, vec3 normal)
{
    // perform perspective divide
    vec3 proj_coords = pos_light_space.xyz / pos_light_space.w;
	
    // transform to [0,1] range
    proj_coords = proj_coords * 0.5 + 0.5;
	
    // get closest depth value from light's perspective (using [0,1] range fragPosLight as coords)
    float closest_depth = texture(depth_map, proj_coords.xy).r;
	
    // get depth of current fragment from light's perspective
    float current_depth = proj_coords.z;
	
    // check whether current frag pos is in shadow
    // bias against shadow acne
    float bias = max(0.001 * (1.0 - dot(normal, light_dir)), 0.001);  
    //float shadow = currentDepth - bias > closestDepth  ? 1.0 : 0.0;

    float shadow = 0.0;
    vec2 texel_size = 1.0 / textureSize(depth_map, 0);
    
    for(int x = -1; x <= 1; ++x)
    {
      for(int y = -1; y <= 1; ++y)
      {
        float pcf_depth = texture(depth_map, proj_coords.xy + vec2(x, y) * texel_size).r;
        shadow += step(pcf_depth, current_depth - bias);
      }    
    }
    
    shadow /= 9.0;
    
    if(proj_coords.z > 1.0) {
        shadow = 0.0;
    }
	
    return shadow;
}

//https://blog.selfshadow.com/publications/s2013-shading-course/karis/s2013_pbs_epic_notes_v2.pdf
float brdf_cook_torrance(vec3 v, vec3 n, float nDl, vec3 h, float roughness, float metallic) {
  float nDh = max(dot(n, h), 0);
  float nDv = max(dot(n, v), 0);
  float vDh = max(dot(v, h), 0);
  
  // geometric
  float geo_num = 2.0 * nDh;
  float geo_denom = 1.0 / vDh;
  float geo_b = (geo_num * nDv) * geo_denom;
  float geo_c = (geo_num * nDl) * geo_denom;
  float geo = min(min(geo_b, geo_c), 1.0);

  // microfacet
  float rough2 = roughness * roughness;
  float rough_a = 1.0 / (4.0 * rough2 * pow(nDh, 4.0));
  float rough_b = nDh * nDh - 1.0;
  float rough_c = rough2 * nDh * nDh;
  float rough = rough_a * exp(rough_b/rough_c);
  /*
  float r2 = roughness * roughness;
  float nDh2 = nDh * nDh;
  float a = 1.0/(3.14159*r2*pow(nDh, 4));
  float b = exp((nDh2 - 1) / r2 * nDh2);
  rough = a*b;*/
  
  // fresnel
  float fresnel = pow(1.0 - vDh, 5.0);
  fresnel *= (1.0 - metallic);
  fresnel += metallic;
  
  return (fresnel * geo * rough) / (4 * nDv * nDl);
}

vec3 environment(vec3 v, vec3 n) {
  vec3 r = reflect(-v, n);
  return texture(skybox, r).rgb;
}

const vec2 terrain_scale = vec2(0.1, 0.1);

vec3 pbr(vec3 v, vec3 n, vec3 l, vec3 h) {
  float metallic = texture(texture_metallic, vert_uv).r;
  float roughness = texture(texture_roughness, vert_uv).r;
  
  float nDl = max(dot(n, l), 0.0);
  
  vec3 material_color = texture(texture_color, vert_uv).rgb * base_color.rgb;
  
  if (model_type > 0.0f) {
    vec3 top_color = texture(texture_ao, vert_world_position.xz*terrain_scale).rgb;
    material_color = mix(material_color, top_color, clamp(pow(dot(normalize(vert_normal), vec3(0,1,0))+0.2f,16.0), 0.0, 1.0));
    metallic = 0.0;
    roughness = 0.0;
  }
  
  vec3 brdf_specular = vec3(0);
  if (nDl > 0.0f) {
    brdf_specular = vec3(brdf_cook_torrance(v, n, nDl, h, roughness, metallic));
  }
  
  vec3 diffuse = mix(material_color, material_color * environment(v, n), metallic*max_metallic);
  vec3 specular = max(brdf_specular * light_color, vec3(0));

  return (diffuse + specular) * nDl;
}

void main() {
  
  vec3 view_dir = normalize(camera_position - vert_world_position);
  vec3 half_dir = normalize(light_dir + view_dir);
  
  
  // normal calculation
  vec3 normal_bump = texture(texture_normal, vert_uv).rgb;         // obtain normal from normal map in range [0,1]
  if (model_type > 0.0f) {
    vec3 x = texture(texture_normal, vert_world_position.zy*terrain_scale).rgb;
    vec3 y = texture(texture_normal, vert_world_position.zx*terrain_scale).rgb;
    vec3 z = texture(texture_normal, vert_world_position.xy*terrain_scale).rgb;
    
    vec3 blend_normal = clamp(pow(vec3(0,1,0)*1.4,vec3(4.0)), 0.0, 1.0);
    vec3 blended_t = z;
    blended_t = mix(blended_t, x, blend_normal.x);
    blended_t = mix(blended_t, y, blend_normal.y);
    
    normal_bump = blended_t;
  }
  
  
  vec3 normal_tangent_space = normalize(normal_bump * 2.0 - 1.0);  // transform normal vector to range [-1,1]
  vec3 normal = normalize(TBN * normal_tangent_space);             // this normal_bump is in tangent space
  
  //normal = normalize(vert_normal);
  
  // shadow calculation
  float shadow = shadow_calculation(vert_light_pos, normal)*shadow_damping;
  
  vec3 ambient_occlusion = mix(texture(texture_ao, vert_uv).rgb,vec3(1),model_type);
  
  vec3 color = pbr(view_dir, normal, light_dir, half_dir);
  
  
  // fog calculation
  float dist = length(camera_position -vert_world_position);
  float fog_factor = 1.0 /exp( (dist * fog.a)* (dist * fog.a));
  fog_factor = clamp( fog_factor, 0.0, 1.0 );
  
  vec3 shaded = color * min(ambient_occlusion,vec3(1.0 - shadow)) + ambient;

  frag_color = vec4(mix(fog.rgb, shaded, fog_factor),1.0);
  
  float brightness = dot(frag_color.rgb, vec3(0.2126, 0.7152, 0.0722));
  glow_color = vec4(0);//mix(vec4(color.rgb, glow), vec4(color.rgb, bloom_damping), step(brightness_threshold, brightness));
}