#version 430

// outputs from vertex shader
in vec3 vert_world_position;
in vec3 vert_normal;
in vec2 vert_uv;
in vec4 vert_light_pos;
in vec3 light_direction;
in mat3 TBN;

layout(location = 31) uniform sampler2D texture_color;
layout(location = 32) uniform sampler2D texture_normal;
layout(location = 33) uniform sampler2D texture_metallic;
layout(location = 34) uniform sampler2D texture_roughness;
layout(location = 35) uniform sampler2D texture_ao;

layout(location = 36) uniform sampler2D depth_map;

layout(location = 21) uniform float glow;

layout(location = 26) uniform vec3 light_dir;
layout(location = 27) uniform vec3 light_color;
layout(location = 28) uniform vec3 camera_position;

const vec3 ambient = vec3(0.1);
const vec3 specular_color = vec3(1.0, 1.0, 1.0);
const float shininess = 16.0;
const float brightness_threshold = 1.3;
const float bloom_damping = 0.25;
const float shadow_damping = 0.6;

layout(location = 0) out vec4 frag_color;
layout(location = 1) out vec4 glow_color;
layout(location = 2) out vec4 normal_color;

float shadow_calculation(vec4 pos_light_space, vec3 normal)
{
    // perform perspective divide
    vec3 proj_coords = pos_light_space.xyz / pos_light_space.w;
	
    // transform to [0,1] range
    proj_coords = proj_coords * 0.5 + 0.5;
	
    // get closest depth value from light's perspective (using [0,1] range fragPosLight as coords)
    float closest_depth = texture(depth_map, proj_coords.xy).r;
	
    // get depth of current fragment from light's perspective
    float current_depth = proj_coords.z;
	
    // check whether current frag pos is in shadow
    // bias against shadow acne
    float bias = max(0.001 * (1.0 - dot(normal, light_dir)), 0.001);  
    //float shadow = currentDepth - bias > closestDepth  ? 1.0 : 0.0;

    float shadow = 0.0;
    vec2 texel_size = 1.0 / textureSize(depth_map, 0);
    
    for(int x = -1; x <= 1; ++x)
    {
      for(int y = -1; y <= 1; ++y)
      {
        float pcf_depth = texture(depth_map, proj_coords.xy + vec2(x, y) * texel_size).r;
        shadow += step(pcf_depth, current_depth - bias);
      }    
    }
    
    shadow /= 9.0;
    
    if(proj_coords.z > 1.0) {
        shadow = 0.0;
    }
	
    return shadow;
}


void main() {
  
  
  // obtain normal from normal map in range [0,1]
  vec3 normal_bump = texture(texture_normal, vert_uv).rgb;
  // transform normal vector to range [-1,1]
  vec3 normal_tangent_space = normalize(normal_bump * 2.0 - 1.0);  // this normal_bump is in tangent space
  vec3 normal = normalize(TBN * normal_tangent_space);
  
  float diffuse = max(dot(light_dir, normal), 0.0);
  float specular = 0.0;
  
  float shadow = shadow_calculation(vert_light_pos, normal)*shadow_damping;
  
  if (diffuse > 0.0) {
    vec3 view_dir = normalize(camera_position - vert_world_position);
    vec3 half_dir = normalize(light_dir + view_dir);
    specular = pow(max(dot(half_dir, normal), 0.0), shininess);
  }
  
  vec4 tex_color = texture(texture_color, vert_uv);
  vec3 light = (1.0 - shadow) * ((diffuse * light_color) + (specular * specular_color)) + ambient;
  
  vec3 color = light * tex_color.rgb;
  
  frag_color = vec4(color,1.0)*texture(texture_ao, vert_uv);
  
  float brightness = dot(frag_color.rgb, vec3(0.2126, 0.7152, 0.0722));
  glow_color = mix(vec4(color.rgb, glow), vec4(color.rgb, bloom_damping), step(brightness_threshold, brightness));
}